#!/usr/bin/python3
# -*- coding: utf-8 -*-
from math import log10
import os
import re
import json
import shutil
import sqlite3
import logging
import datetime
from gettext import gettext as _

import apt_pkg
from ..backend import InstallBackend
DB_FILE = os.path.join("/var/cache/kylin-system-updater/kylin-system-updater.db")
UMDB_FILE = os.path.join("/var/cache/kylin-system-updater/kylin-system-updater.db")
INSTALLED_LIST = [{"item": "errorcode", "type": "int", "default": "0"}]
DISPALY_LIST = []

class Sqlite3Server(object):
    def __init__(self, window_main):
        self.connect = None
        self.window_main = window_main
        logging.info(_("Init Sqlite3Server..."))
        self.init_sqlit()
        
    # 初始化连接数据库
    def init_sqlit(self):
        logging.info(_("Initialize the connection to the database ..."))
        try: 
            if os.path.isfile(DB_FILE):
                self.connect = sqlite3.connect(DB_FILE, check_same_thread=False)
                self.cursor = self.connect.cursor()
                self.insert_new_field()
            else:
                if not os.path.isdir(os.path.dirname(DB_FILE)):
                    os.makedirs(os.path.dirname(DB_FILE))
                shutil.copy("/usr/share/kylin-system-updater/kylin-system-updater.db", os.path.dirname(DB_FILE))
                self.connect = sqlite3.connect(DB_FILE, check_same_thread=False)
                self.cursor = self.connect.cursor()
                self.insert_new_field()
        except Exception as e:
            logging.error(_("Failed to initialize the database: %s"), str(e))
            return False
        logging.info(_("Connecting to the database successfully."))
        return True

    # 数据库表格中动态增加新的字段用于扩展
    def insert_new_field(self):
        if len(INSTALLED_LIST) == 0 and len(DISPALY_LIST) == 0:
            return
        self.cursor.execute("select sql from sqlite_master where name='installed'")
        installed_sql = self.cursor.fetchone()[0]
        pattern = re.compile(r'\"\w+\"')
        installed_sql_list = pattern.findall(installed_sql)
        for value in INSTALLED_LIST:
            for field in installed_sql_list:
                if value["item"] == str(field).strip("\""):
                    break
                elif field == installed_sql_list[len(installed_sql_list) - 1]:
                    try:
                        if value["default"] != "":
                            sql = 'alter table installed add column "' + value["item"] + '" ' + value["type"] \
                                  + ' default ' + str(value["default"])
                        else:
                            sql = 'alter table installed add column "' + value["item"] + '" ' + value["type"]
                        self.cursor.execute(sql)
                        logging.info(_("installed table insert new field: %s"), value["item"])
                    except:
                        logging.error(_("installed table failed to insert a new field:"), value["item"], exc_info=True)

        self.cursor.execute("select sql from sqlite_master where name='display'")
        display_sql = self.cursor.fetchone()[0]
        pattern = re.compile(r'\"\w+\"')
        display_sql_list = pattern.findall(display_sql)
        for value in DISPALY_LIST:
            for field in display_sql_list:
                if value["item"] == str(field).strip("\""):
                    break
                elif field == display_sql_list[len(display_sql_list) - 1]:
                    try:
                        if value["default"] != "":
                            sql = 'alter table display add column "' + value["item"] + '" ' + value["type"] \
                                  + ' default ' + str(value["default"])
                        else:
                            sql = 'alter table installed add column "' + value["item"] + '" ' + value["type"]
                        self.cursor.execute(sql)
                        logging.info(_("display table insert new field： %s"), value["item"])
                    except:
                        logging.error(_("display table failed to insert a new field： %s"), value["item"], exc_info=True)

    # 写入数据到installed表中
    def insert_into_installed(self, *args, **kwargs):
        self.cursor.execute(
            "insert into installed (appname, version, time, description, icon, statue, keyword, errorcode) values(?,"
            "?,?,?,?,?,?,?)",
            (args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7]))
        self.connect.commit()

    # 写入数据到display表中
    def insert_into_display(self, *args, **kwargs):
        try:
            sql = "update display set " + args[0] + "='" + args[1] + "' where id = 1"
            self.cursor.execute(sql)
            self.connect.commit()
        except Exception as e:
            logging.error("Insert error: %s.", str(e))
        logging.info(_("Data insertion complete ."))

    # 读出display表中数据
    def select_from_display(self, *args, **kwargs):
        try:
            sql = "select "+args[0]+" from display"
            self.cursor.execute(sql)
            self.connect.commit()
            return str(self.cursor.fetchone()[0])
        except Exception as e:
            logging.error("Insert error: %s.", str(e))
        logging.info(_("Data insertion complete ."))

    # 写入updateinfos表中
    def insert_into_updateinfo(self, *args, **kwargs):
        logging.info(_("Inserting data into the database... "))
        try:
            self.cursor.execute(
                "insert into updateinfos (appname, version, description, date, status, keyword, errorcode) values(?,"
                "?,?,?,?,?,?)",
                (args[0], args[1], args[2], args[3], args[4], args[5], args[6]))
            self.connect.commit()
        except Exception as e:
            logging.error("Insert error: %s.", str(e))
        logging.info(_("Data insertion complete ."))
    
    # 接收更新列表与信息，生成数据并插入数据库中
    def insert_info(self, mode, pkg_list=[], pkg_group=[], adjust_pkg=[], success = False, error_string = '', error_desc = ''):
        errstr = error_string + " " + error_desc
        status = " "
        UpdateInfos  = {}
        InstallInfos = {}
        time = datetime.datetime.now()
        timestr = datetime.datetime.strftime(time, "%Y-%m-%d %H:%M:%S")

        # 更新列表空，无更新
        if not pkg_list and not pkg_group and mode != InstallBackend.MODE_UPGRADE_SYSTEM:
            logging.info("There is no update.")
            return True

        if success:
            status = 'success'
        else:
            status = 'failed'

        # 判断更新方式
        if mode == InstallBackend.MODE_UPGRADE_PARTIAL: # 部分更新
            pkg_adj = ""
            # 判断更新包为单包或更新组
            if pkg_group:
            # 更新组
                pkgname = pkg_group.pop(0)
                pkgversion,pkgdescription = self.GetGroupmsg(pkgname)
                #更新信息update-infos
                UpdateInfos.update({"packageName":str(pkgname)})
                UpdateInfos.update({"source":"Kylin Syetm Updater"})
                UpdateInfos.update({"status":status})
                UpdateInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(UpdateInfos.copy())
                UpdateInfos.clear()
                # self.window_main.collector.UpdateMsg("UpdateInfos", json_file)
                #安装信息install-infos
                InstallInfos.update({"packageName":str(pkgname)})
                if pkgname in self.window_main.update_list.local_upgrade_data.versoin_pkgs['groups_upgrade'].keys():
                    InstallInfos.update({"old_version":str(self.window_main.update_list.local_upgrade_data.versoin_pkgs['groups_upgrade'][pkgname])})
                else:
                    InstallInfos.update({"old_version":'UnKnown'})
                InstallInfos.update({"new_version":str(pkgversion)})
                InstallInfos.update({"status":status})
                InstallInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(InstallInfos.copy())
                # self.window_main.collector.UpdateMsg("InstallInfos", json_file)
            elif pkg_list:
                # 单包更新 # 获取单包数据插入数据库
                pkgname = pkg_list.pop(0)
                for adj in adjust_pkg:
                    if pkgname in adj:
                    # 该部分升级的单包为调整版本，与候选版本不一致
                        pkg_adj = adj
                        break
                if pkg_adj: # 有调整的情况
                    try:
                        pkg =  self.window_main.cache[pkg_adj.split("=")[0]]
                        for ver in pkg.versions:
                            if ver.version == pkg_adj.split("=")[1]:
                                pkg_inst_ver = ver
                                break
                        pkgname = pkg_adj.split("=")[0]
                        pkgversion = str(pkg_inst_ver.source_version)
                        pkgdescription = str(pkg_inst_ver.description)
                    except Exception as e:
                        logging.error(_("%s could not be detected in the source because the source was changed or for other reasons."), \
                            str(pkgname))
                        logging.error(str(e))
                else: # 没有调整的情况
                    try:
                        pkg =  self.window_main.cache[pkgname]
                        pkgversion = str(pkg.candidate.version)
                        pkgdescription = str(pkg.candidate.raw_description)
                    except Exception as e:
                        logging.error(str(e))

                #更新信息update-infos
                UpdateInfos.update({"packageName":str(pkgname)})
                UpdateInfos.update({"source":"Kylin Syetm Updater"})
                UpdateInfos.update({"status":status})
                UpdateInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(UpdateInfos.copy())
                UpdateInfos.clear()
                # self.window_main.collector.UpdateMsg("UpdateInfos", json_file)
                #安装信息install-infos
                InstallInfos.update({"packageName":str(pkgname)})
                if pkgname in self.window_main.update_list.local_upgrade_data.versoin_pkgs['single_upgrade'].keys():
                    InstallInfos.update({"old_version":str(self.window_main.update_list.local_upgrade_data.versoin_pkgs['single_upgrade'][pkgname])})
                else:
                    InstallInfos.update({"old_version":'UnKnown'})
                InstallInfos.update({"new_version":str(pkgversion)})
                InstallInfos.update({"status":status})
                InstallInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(InstallInfos.copy())
                # self.window_main.collector.UpdateMsg("InstallInfos", json_file)
                
            try:
                self.insert_into_updateinfo(pkgname, pkgversion, pkgdescription, timestr, status, "1", errstr)
                # FIXME: 发送插入数据库成功的信号local_upgrade_list
                self.window_main.dbusController.UpdateSqlitSingle(pkgname, timestr)
            # 数据库文件被删除或者新增字段导致需要重新初始化数据库再写入
            except Exception as e:
                self.init_sqlit()
                self.insert_into_updateinfo(pkgname, pkgversion, pkgdescription, timestr, status, "1", errstr)
                # FIXME: 这里也需要, 发送插入数据库成功的信号 
                self.window_main.dbusController.UpdateSqlitSingle(pkgname, timestr)
        elif mode == InstallBackend.MODE_UPGRADE_ALL: # 系统全部升级
            # # insert signal deb first
            for i in pkg_list:
                try:
                    pkg =  self.window_main.cache[i]
                except Exception as e:
                    logging.error(_("%s could not be detected in the source because the source was changed or for other reasons."), \
                            str(i))
                    continue
                if not pkg:
                    continue 
                pkgversion = str(pkg.candidate.version)
                pkgdescription = str(pkg.candidate.raw_description)

                #更新信息update-infos
                UpdateInfos.update({"packageName":str(pkg.name)})
                UpdateInfos.update({"source":"Kylin Syetm Updater"})
                UpdateInfos.update({"status":status})
                UpdateInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(UpdateInfos.copy())
                UpdateInfos.clear()
                # self.window_main.collector.UpdateMsg("UpdateInfos", json_file)
                #安装信息install-infos
                InstallInfos.update({"packageName":str(pkg.name)})
                if pkg.name in self.window_main.update_list.local_upgrade_data.versoin_pkgs['groups_upgrade'].keys():
                    InstallInfos.update({"old_version":str(self.window_main.update_list.local_upgrade_data.versoin_pkgs['groups_upgrade'][pkg.name])})
                else:
                    InstallInfos.update({"old_version":'UnKnown'})
                InstallInfos.update({"new_version":str(pkgversion)})
                InstallInfos.update({"status":status})
                InstallInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(InstallInfos.copy())
                # self.window_main.collector.UpdateMsg("InstallInfos", json_file)

                try:
                    self.insert_into_updateinfo(str(i), pkgversion, pkgdescription, timestr, status, "1", errstr)
                    self.window_main.dbusController.UpdateSqlitSingle(str(i), timestr)
                # 数据库文件被删除或者新增字段导致需要重新初始化数据库再写入
                except Exception as e:
                    self.init_sqlit()
                    self.insert_into_updateinfo(str(i), pkgversion, pkgdescription, timestr, status, "1", errstr)
                    self.window_main.dbusController.UpdateSqlitSingle(str(i), timestr)
            # insert group deb next
            for i in pkg_group:
                # FIXME: 获取组信息
                pkgversion,pkgdescription = self.GetGroupmsg(i)
                #更新信息update-infos
                UpdateInfos.update({"packageName":str(i)})
                UpdateInfos.update({"source":"Kylin Syetm Updater"})
                UpdateInfos.update({"status":status})
                UpdateInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(UpdateInfos.copy())
                UpdateInfos.clear()
                # self.window_main.collector.UpdateMsg("UpdateInfos", json_file)
                #安装信息install-infos
                InstallInfos.update({"packageName":str(i)})
                if i in self.window_main.update_list.local_upgrade_data.versoin_pkgs['groups_upgrade'].keys():
                    InstallInfos.update({"old_version":str(self.window_main.update_list.local_upgrade_data.versoin_pkgs['groups_upgrade'][i])})
                else:
                    InstallInfos.update({"old_version":'UnKnown'})
                InstallInfos.update({"new_version":str(pkgversion)})
                InstallInfos.update({"status":status})
                InstallInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(InstallInfos.copy())
                # self.window_main.collector.UpdateMsg("InstallInfos", json_file)
                try:
                    self.insert_into_updateinfo(str(i), pkgversion, pkgdescription, timestr, status, "1", errstr)
                    group_list = []
                    self.window_main.dbusController.UpdateSqlitSingle(str(i), timestr)
                # 数据库文件被删除或者新增字段导致需要重新初始化数据库再写入
                except Exception as e:
                    self.init_sqlit()
                    self.insert_into_updateinfo(str(i), pkgversion, pkgdescription, timestr, status, "1", errstr)
                    self.window_main.dbusController.UpdateSqlitSingle(str(i), timestr)
        elif mode == InstallBackend.MODE_UPGRADE_SYSTEM: # 全盘升级
            self.insert_into_updateinfo(str("Upgrade System"), "", "This is a complete system upgrade, equivalent to the implementation of apt dist-upgrade", timestr, status, "1", errstr)
            self.window_main.dbusController.UpdateSqlitSingle(str("Upgrade System"), timestr)
        else:
            logging.warning("Cache is None.")

    # 获取group信息    
    def GetGroupmsg(self, appname):
        INPUT_CONFIG_PATH = apt_pkg.config.find_dir("Kylin-system-updater::InputConfigDir",
                                '/usr/share/kylin-update-desktop-config/data')
        jsonfile = appname+".json"
        files = os.listdir(INPUT_CONFIG_PATH) #获取文件夹中所有文件
        if jsonfile in files: # 存在
            # 读取组JSON文件
            with open(INPUT_CONFIG_PATH+"/"+jsonfile, "r") as f:
                try :
                    data = json.load(f)
                except json.JSONDecodeError as e:
                    logging.error(str(e))
            try:
                version = data['version']
                tmpdescription = data['description']
            except Exception as e:
                logging.error(str(e))
            if "zh_CN" in tmpdescription and "en_US" in tmpdescription:
                description = tmpdescription["zh_CN"] + ": " + tmpdescription["en_US"]
            return (version,description)
        else: #  不存在
            return (None, None)

    def refreshpkglist(self):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove  = []

        for pkg in self.window_main.cache:
                try:
                    if pkg.marked_install:
                        pkgs_install.append(pkg.name)
                    if pkg.marked_upgrade:
                        pkgs_upgrade.append(pkg.name)
                    elif pkg.marked_delete:
                        pkgs_remove.append(pkg.name)
                except KeyError:
                # pkg missing from fresh_cache can't be modified
                    pass
        return pkgs_install,pkgs_upgrade,pkgs_remove
    
    #查找数据库
    def find_msg_from_datebase(self, table, field, action = 'check', cid = 0):
        # 查询数据
        try:
            sql = "select "+field+" from "+table
            self.cursor.execute(sql)
            update_count = self.cursor.fetchone()[0]
            logging.info("%d history updates detected.", update_count)
        except Exception as e:
            logging.error("Check update error: %s", str(e))

def listtojsonstr(lists):
    import json
    jsonfile = json.dumps(lists)
    return jsonfile